/**
 * @file main.c
 * @brief Este projeto demonstra o uso do modulo RTC, exibindo os segundos no LCD
 * @author Wu Shin Ting
 * @date 25/10/2020
 */

#include "string.h"
#include "derivative.h" /* include peripheral declarations */

/**
 * @brief Dois estados booleanos
 */
typedef enum estado_tag {
	DESABILITA,     /**< falso/apaga/desativa/liga */
	HABILITA,      /**< verdadeiro/acende/ativa/fecha */
} estado_type;

#define GPIO_PIN(x) (1<<x)

estado_type alarm;
//==================================
// Funcoes utilitarias
//==================================
/**
 * @brief atraso em numero de interacoes <i>
 * @param[in] i numero de iteracoes
 */
void delay (uint32_t i)
{
	__asm("outerLoop:;"
				"sub %0, %0, #1;"
				"mov r2, #11;"
			"innerLoop:;"
				"sub r2, r2, #1;"
				"cmp r2, #0;"
				"bne innerLoop;"
				"cmp %0, #0;"
				"bne outerLoop;"
			: 
			: "r" (i)
			);
}

char *ttoa (unsigned int seconds, char *string)
{
	unsigned int hh, mm, ss;
	
	ss = seconds%60;
	
	mm = (seconds/60)%60;
	
	hh = seconds/3600;
	
	string[2] = string[5] = ':';    // inserir os espacadores
	string[8] = '\0';				// terminador

	if (hh > 23) {
		// horario invalido: FF:FF:FF
		string[0] = string[1] = string[3] = string[4] = string[6] = string[7] = 'F';
	}
	
	string[0] = (hh < 10)? '0': (hh/10)+'0';
	string[1] = hh%10+'0';
	
	string[3] = (mm < 10)? '0': (mm/10)+'0';
	string[4] = mm%10+'0';
	
	string[6] = (ss < 10)? '0': (ss/10)+'0';
	string[7] = ss%10+'0';	
	
	return string;
}

//==================================
// LCD
//==================================
/**
 * @brief Funcao do byte
 */
typedef enum lcd_RS_tag {
	COMANDO,  /**< Comando */
	DADO	  /**< Dado */
} lcd_RS_type;

/**
 * @brief Inicializa os pinos conectados ao latch 74573 onde estao conectados os 8 leds vermelhos
 */
void GPIO_initLCD(){
	SIM_SCGC5 |= SIM_SCGC5_PORTC_MASK ;     // Habilitar o clock da PORTA C

	// Configura os pinos PORTC[10:0] como GPIO de saida 
	PORTC_PCR0 |= PORT_PCR_MUX(0x1);       // D0-D7 dos dados (GPIO)
	PORTC_PCR1 |= PORT_PCR_MUX(0x1);
	PORTC_PCR2 |= PORT_PCR_MUX(0x1);
	PORTC_PCR3 |= PORT_PCR_MUX(0x1);
	PORTC_PCR4 |= PORT_PCR_MUX(0x1);
	PORTC_PCR5 |= PORT_PCR_MUX(0x1);
	PORTC_PCR6 |= PORT_PCR_MUX(0x1);
	PORTC_PCR7 |= PORT_PCR_MUX(0x1);
	
	PORTC_PCR8 |= PORT_PCR_MUX(0x1);       // RS do LCD
	PORTC_PCR9 |= PORT_PCR_MUX(0x1);       // E do LCD

	GPIOC_PDDR |= GPIO_PDDR_PDD(GPIO_PIN(9)|GPIO_PIN(8) |GPIO_PIN(7) | 
			GPIO_PIN(6) | GPIO_PIN(5) | GPIO_PIN(4) | GPIO_PIN(3) | 
			GPIO_PIN(2) | GPIO_PIN(1) | GPIO_PIN(0));       // Configurar como saidas
	GPIOC_PDOR &= ~GPIO_PDOR_PDO(GPIO_PIN(9) |GPIO_PIN(8)|GPIO_PIN(7) | 
			GPIO_PIN(6) | GPIO_PIN(5) | GPIO_PIN(4) | GPIO_PIN(3) | 
			GPIO_PIN(2) | GPIO_PIN(1) | GPIO_PIN(0)); // Inicializar os pinos com 0
}

/**
 * @brief Escrever nos pinos de dados um caractere de 8 bits
 * @param[in] c caractere
 */
void GPIO_setByte(char c) {
	// colocar os dados nos pinos de dados DO-D7
	GPIOC_PDOR &= 0xFFFFFF00; 		// zerar o byte menos significativo
	GPIOC_PDOR |= (unsigned int) c; // setas os bits em 1 no byte menos significativo	
}

/**
 * @brief Gerar o pulso de Enable de ~450ns
 */
void GPIO_enableE()
{
	// gerar um pulso  (~ 450 ns)
	GPIOC_PSOR = GPIO_PIN (9);		// setar em 1
	delay(1);
	GPIOC_PCOR = GPIO_PIN (9);		// resetar em 0
}

/**
 * @brief Setar o tipo de transferencia
 * @param[in] i tipo
 */
void GPIO_setRS (lcd_RS_type i) {
	if(i == COMANDO) 
		GPIOC_PCOR = GPIO_PIN(8);     // Seta o LCD no modo de comando		
	else if (i == DADO) 
		GPIOC_PSOR = GPIO_PIN(8);    // Seta o LCD no modo de dados
}

/**
 * @brief Transferir um byte ao LCD
 * @param[in] c byte
 * @param[in] t tempo de processamento em multiplos de 2 us
 */
void escreveLCD(char c, uint32_t t)
{
	GPIO_setByte(c);
	GPIO_enableE();
	delay(t);
}

/**
 * @brief Inicializar o LCD
 */
void initLCD () {
	//espera por mais de 30ms = 30000us/2 = 15000
	delay(15000);
	
	GPIO_setRS(COMANDO);
	escreveLCD(0x38, 20);			//Function Set: 39us/2 -> 20
	escreveLCD(0x0C, 20);		    //Display ON/OFF Control: 39us/2 -> 20
	escreveLCD(0x01, 765);			//Display Clear: 1530us/2 -> 765
	escreveLCD(0x06, 20);			//Entry mode set: 39us/2 -> 20
}

/**
 * @brief Limpe o display
 */
void limpeDisplayLCD () {
	GPIO_setRS(COMANDO);
	escreveLCD(0x01, 765);			//Display Clear: 1530us/2 -> 765
}

/**
 * @brief Escrever uma mensagem str no LCD a partir do endereco especificado
 * @param[in] end endereco DDRAM
 * @param[in] str mensagem
 */
void escreveMensagem (uint8_t end, char* str)
{
	//Posicione o cursor no end
	uint8_t tmp = 0b10000000 | end; // seta o bit 7 em 1
	GPIO_setRS(COMANDO);			// comando
	escreveLCD(tmp, 20);			// seta o endereco do cursor

	//Chaveia para modo DADO
	GPIO_setRS (DADO);
	while (*str) {
		escreveLCD(*str, 20);
		str++;
	}
}

/**
 * @brief Escrever um bitmap na memoria CGRAM a partir do endereco especificado
 * @param[in] end endereco CGRAM
 * @param[in] str bitmap
 */
void escreveBitmap (uint8_t end, char* bitmap)
{
	//Posicione o cursor no end
	uint8_t tmp = 0b01000000 | (end*8); // seta o bit 7 em 0 e bit 6 em 1
	GPIO_setRS(COMANDO);				// comando
	escreveLCD(tmp, 20);				// seta o endereco do cursor

	//Chaveia para modo DADO
	GPIO_setRS (DADO);
	uint8_t i;
	for (i=0; i< 8; i++) {
		escreveLCD(bitmap[i], 20);
	}
}

//==================================
// RTC
//==================================
/**
 * @brief Inicializar o modulo RTC
 */
void RTC_init()
{
	SIM_SCGC6 |= SIM_SCGC6_RTC_MASK;	// habilitar sinal CLK para RTC
	
	SIM_SOPT1 |= SIM_SOPT1_OSC32KSEL(0b11);  // setar LPO 1kHz como fonte de ERCLK32K
	
	//Ao inves do prescaler (RTC_TPR) incrementar a 32.768kHz, incrementara a 1kHz 

	RTC_CR |= RTC_CR_SWR_MASK;			// resetar os registradores do RTC por software
										// limpar o contador RTC_TSR e flags RTC_SR_TOF e RTC_SR_TIF
	RTC_CR &= ~RTC_CR_SWR_MASK;			// necessario resetar explicitamente SWR 
	
    RTC_TCR = RTC_TCR_CIR(0x00) |		// ajustar o erro de frequencia (ppm) em cada segundo
              RTC_TCR_TCR(0x00);        // 1s = 32.768 ticks (sem ajuste)
    
    RTC_TPR = 0; 			// resetar prescaler (2^15 - 1)
    RTC_TSR = 0;			// resetar o contador para que a bandeira 
    						// TIF seja baixada.
    
	RTC_IER &= ~(RTC_IER_TAIE_MASK |	// desabilitar a interrupcao por alarme
				RTC_IER_TOIE_MASK  |	// desabilitar a interrupcao por overflow
				RTC_IER_TSIE_MASK);    	// desabilitar a interrupcao por segundo

	RTC_SR |= RTC_SR_TCE_MASK;			// habilita incremento de Prescaler
}

/**
 * @brief Habilitar o tratamento de interrupcao por alarme no NVIC
 */
void RTC_enableNVICAlarm (char priority) {

    NVIC_ISER |= 1 << (36-16);            ///< habilitar IRQ 20
    NVIC_ICPR |= 1 << (36-16);
    NVIC_IPR5 |= NVIC_IP_PRI_20(priority << 6);   // 20/4
}

/**
 * @brief Habilitar o tratamento de interrupcao por segundo no NVIC
 */
void RTC_enableNVICSecond (char priority) {

    NVIC_ISER |= 1 << (37-16);            ///< habilitar IRQ 21
    NVIC_ICPR |= 1 << (37-16);
    NVIC_IPR5 |= NVIC_IP_PRI_21(priority << 6);   // 21/4
}

/**
 * @brief Habilitar interrupcao por alarme
 */
void RTC_enableInterruptAlarm (void) {
	RTC_IER |= RTC_IER_TAIE_MASK;    	// habilitar a interrupcao por segundo
}

/**
 * @brief Desabilitar a interrupcao por alarme
 */
void RTC_disableInterruptAlarm(void) {
	RTC_IER &= ~RTC_IER_TAIE_MASK;    // desabilitar a interrupcao por segundo
}

/**
 * @brief Habilitar interrupcao por segundo
 */
void RTC_enableInterruptSecond (void) {
	RTC_IER |= RTC_IER_TSIE_MASK;    	// habilitar a interrupcao por segundo
}

/**
 * @brief Desabilitar a interrupcao por segundo
 */
void RTC_disableInterruptSecond(void) {
	RTC_IER &= ~RTC_IER_TSIE_MASK;    // desabilitar a interrupcao por segundo
}

/**
 * @brief Tratar o evento de alarme, overflow e tempo invalido do modulo RTC
 */
void RTC_Alarm_IRQHandler(void) {
	static unsigned int i=1;
	if (RTC_SR & RTC_SR_TAF_MASK) {
		// evento Alarme
		alarm = HABILITA;
		RTC_TAR = (2*i++); // escrita no registrador de alarme baixa flag de solicitacao
		// soar depois do primeiro intervalo de alarme ...
	}
	if (RTC_SR & RTC_SR_TIF_MASK) {
		// evento tempo invalido
		RTC_SR &= ~RTC_SR_TCE_MASK;			// resetar o prescaler e o contador depois de desabilitar RTC

		RTC_TPR = 0;
		RTC_TSR = 0;		
		RTC_SR |= RTC_SR_TCE_MASK;			// habilitar RTC
	}
}

/**
 * @brief Tratar o evento de asgundo do modulo RTC
 */
void RTC_Seconds_IRQHandler(void) {
	// A base de tempo do nosso relogio eh 1/32768 menor. Portanto,
	// o tick de um segundo eh gerado somente a cada intervalo de 32768 segundos.
}

/**
 * @brief Ler o valor atual do contador de segundos do RTC
 * @param[out] seconds medida de tempo em segundos
 */
void RTC_getTime (unsigned int *seconds) {
	*seconds = (RTC_TSR*32768)/1000+(RTC_TPR/1000);    // reajustar valor para segundos
}

/**
 * @brief Setar o valor do contador de segundos do RTC
 * @param[in] seconds valor de tempo em segundos
 */
void RTC_setTime (unsigned int seconds) {
	RTC_SR &= ~RTC_SR_TCE_MASK;			// desabilita o modulo para atualizar prescaler
										// e contador
	RTC_TPR = (seconds*1000)%32768;     // reajustar segundos para a frequencia 1kHz
	RTC_TSR = (seconds*1000)/32768;		// reajustar segundos para a frequencia 1kHz
	RTC_SR |= RTC_SR_TCE_MASK;			// habilitar o modulo
}

/**
 * @brief Setar o valor do alarme do RTC 
 * @param[in] seconds valor de alarme em segundos
 */
void RTC_setAlarm (unsigned int seconds) {
	RTC_TAR = (seconds*1000)/32768;		// reajustar segundos para a frequencia 1kHz
}

/**
 * @brief Fluxo de controle principal
 */


int main(void)
{
	unsigned int seconds_back, seconds;
	char result[20];
	
	GPIO_initLCD();					// inicializar os pinos de controle do LCD
	initLCD();						// inicializar o LCD
	
	RTC_init();						// configurar os registradores do RTC
	RTC_enableNVICAlarm (0);		// prioridade maxima dentro do grupo 5 de eventos
	// Inicializar o tempo em 0
	RTC_setTime (0);
	escreveMensagem(0x00, "0");
	RTC_setAlarm (60);
	RTC_enableInterruptAlarm ();

	// Inicializar iteracoes
	RTC_getTime (&seconds);
	ttoa (seconds, result);
	escreveMensagem(0x00, result);
	seconds_back = seconds;
	
	alarm = DESABILITA;
	
	for(;;) {
		if (alarm == HABILITA) {
			char vezes = 0;
			for (vezes = 0; vezes < 5; vezes++) {
				escreveMensagem(0x40, "ALARME!!!");
				delay(125000);
				escreveMensagem(0x40, "         ");
				delay(125000);				
			}
			alarm = DESABILITA;
		}
		RTC_getTime (&seconds);
		if (seconds > seconds_back) {
			ttoa (seconds, result);
			escreveMensagem(0x00, result);
			seconds_back = seconds;
		}		
	}
	
	return 0;
}
